// stargazer - A Gemini Server
// Copyright (C) 2021 Sashanoraa <sasha@noraa.gay>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

mod cgi;
mod cli;
mod config;
mod dir_list;
mod error;
mod get_file;
mod logger;
mod router;
mod tls;

use crate::error::{bad_req, ErrorConv, GemError, Result};
use async_channel::{Receiver, Sender};
use async_executor::Executor;
use async_io::Timer;
use async_net::{SocketAddrV6, TcpListener, TcpStream};
use bstr::ByteSlice;
use cgi::{serve_cgi, serve_scgi};
use futures_lite::*;
use futures_rustls::{server::TlsStream, TlsAcceptor};
use get_file::get_file;
use log::{debug, error};
use once_cell::sync::{Lazy, OnceCell};
use proxy_protocol::ProxyHeader;
use router::{route, Request, Route};
use std::clone::Clone;
use std::convert::TryFrom;
use std::net::SocketAddr;
use std::net::{IpAddr, Ipv4Addr, Ipv6Addr};
use std::panic::catch_unwind;
use std::sync::Arc;
use std::{process::exit, str, thread, time::Duration};

pub static SHUTDOWN: Lazy<(Sender<()>, Receiver<()>)> =
    Lazy::new(|| async_channel::bounded(1));
/// Global executor
pub static EXEC: Lazy<Executor> = Lazy::new(|| {
    // Start executor thread per cpu core
    for n in 1..=CONF.worker_threads {
        thread::Builder::new()
            .name(format!("stargazer-worker-{n}"))
            .spawn(|| loop {
                let _ = catch_unwind(|| {
                    async_io::block_on(EXEC.run(SHUTDOWN.1.recv()))
                });
            })
            .expect("cannot spawn executor thread");
    }

    Executor::new()
});

static CONF_BACKING: OnceCell<config::Config> = OnceCell::new();
pub static CONF: Lazy<config::Config> = Lazy::new(|| {
    CONF_BACKING
        .get()
        .cloned()
        .expect("CONF accessed before it was set")
});

fn main() {
    use error::Context;

    // Parse cli args
    let args = cli::Args::from_args();

    // Set log level based on debug cli flag
    let log_level = match args.debug {
        true => log::LevelFilter::Debug,
        false => log::LevelFilter::Warn,
    };

    // start logger
    logger::init(log_level).expect("Error starting logger");

    // Use dev or regular config based on dev cli flag
    let conf = if args.dev {
        println!("Staring dev server at gemini://localhost/");
        config::dev_config()
    } else {
        debug!("Loading config from: {}", args.config_path.display());
        config::load(&args.config_path)
    };
    let conf = match conf {
        Ok(conf) => conf,
        Err(e) => {
            error!("{:#}", e);
            log::logger().flush();
            exit(1);
        }
    };
    debug!("{:#?}", conf);
    if args.check_config {
        exit(0);
    }
    // Set global config
    CONF_BACKING.set(conf).unwrap();

    if CONF.conn_logging && !args.debug {
        log::set_max_level(log::LevelFilter::Info);
    }

    // Configure tls acceptor for all domain names in config
    let acceptor = TlsAcceptor::from(match tls::load_config(&CONF.routes) {
        Ok(conf) => Arc::new(conf),
        Err(e) => {
            error!("{:#}", e);
            log::logger().flush();
            exit(1);
        }
    });

    // Create a listener task for each interface+port listed in conf
    debug!("Server staring");
    for addr in &CONF.listen {
        let acceptor = acceptor.clone();
        EXEC.spawn(async move {
            // Start tcp listener
            // Program exits with an error if any fail to bind
            let listener = match addr {
                SocketAddr::V4(addr) => match TcpListener::bind(addr).await {
                    Ok(l) => l,
                    Err(e) => {
                        error!("{:#}", e);
                        log::logger().flush();
                        exit(1);
                    }
                },
                SocketAddr::V6(addr) => match create_v6_sock(addr) {
                    Ok(l) => l,
                    Err(e) => {
                        error!("Error creating IPv6 listener: {:#}", e);
                        log::logger().flush();
                        exit(1);
                    }
                },
            };

            let server_port = match listener.local_addr() {
                Ok(addr) => addr.port(),
                Err(e) => {
                    error!("{:#}", e);
                    log::logger().flush();
                    exit(1);
                }
            };

            // Accept connections
            loop {
                let (stream, remote_addr) = match listener
                    .accept()
                    .await
                    .context("Error accepting tcp connection")
                {
                    Ok(ret) => ret,
                    Err(e) => {
                        debug!("{:#}", e);
                        continue;
                    }
                };
                let acceptor = acceptor.clone();

                // Start connection handler task
                // This will time out after 10s
                EXEC.spawn(async move {
                    if CONF.response_timeout == 0 {
                        start_task(stream, acceptor, remote_addr, server_port)
                            .await
                    } else {
                        start_task(stream, acceptor, remote_addr, server_port)
                            .or(async {
                                Timer::after(Duration::from_secs(
                                    CONF.response_timeout,
                                ))
                                .await;
                                debug!(
                                "Request timeout: Total request time exceeded"
                            );
                            })
                            .await
                    }
                })
                .detach();
            }
        })
        .detach();
    }

    async_io::block_on(async {
        if let Err(e) = exit_on_sig().await {
            println!("Error exiting !!!: {:#}", e);
        }
    });
}

fn create_v6_sock(addr: &SocketAddrV6) -> Result<TcpListener> {
    use socket2::{Domain, Socket, Type};
    let socket = Socket::new(Domain::IPV6, Type::STREAM, None)?;
    socket.set_only_v6(true)?;
    let addr = addr.to_owned().into();
    socket.bind(&addr)?;
    socket.listen(128)?;
    let listener: std::net::TcpListener = socket.into();
    Ok(TcpListener::try_from(listener)?)
}

#[cfg(unix)]
async fn exit_on_sig() -> Result<()> {
    use signal_hook::consts::{SIGINT, SIGTERM};
    use signal_hook_async_std::Signals;

    let mut signals = Signals::new([SIGTERM, SIGINT])?;
    signals.next().await;
    SHUTDOWN.0.send(()).await?;
    log::logger().flush();
    std::process::exit(0);
}

/// A Reader which can peek into the stream without changing the cursor.
///
/// This is used with TcpStream and str so parse_proxy_protocol can be tested
trait AsyncPeek {
    async fn peek(&mut self, buf: &mut [u8]) -> io::Result<usize>;
}

impl<'a> AsyncPeek for &'a [u8] {
    async fn peek(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        let to_read = buf.len().min(self.len());
        buf[..to_read].copy_from_slice(&self[..to_read]);
        Ok(to_read)
    }
}

impl AsyncPeek for TcpStream {
    async fn peek(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        TcpStream::peek(self, buf).await
    }
}

/// Parse the proxy protocol given a peek buffer and a stream.
///
/// It is separated out from handle_proxy_protocol for testing purposes.
async fn parse_proxy_protocol<R: AsyncReadExt + AsyncPeek + Unpin>(
    stream: &mut R,
) -> anyhow::Result<Option<ProxyHeader>> {
    let mut peek = [0u8; 12];
    let n = stream.peek(&mut peek).await?;
    let peek = &peek[..n];

    if peek.starts_with(b"PROXY ") {
        // V1 - Read until \r\n
        let mut buf: Vec<u8> = Vec::new();
        loop {
            // Peek in blocks of 64 bytes, and read only until the \n.
            let mut peek = [0u8; 64];
            let n = stream.peek(&mut peek).await?;
            let peek = &mut peek[..n];
            let (found, i) = match peek.find(b"\n") {
                Some(i) => (true, i + 1),
                None => (false, n),
            };
            let peek = &mut peek[..i];
            stream.read_exact(peek).await?;
            buf.extend_from_slice(peek);
            if found {
                break;
            }
        }
        let mut buf = buf.as_slice();

        Ok(Some(proxy_protocol::parse(&mut buf)?))
    } else if peek.starts_with(&[
        0x0D, 0x0A, 0x0D, 0x0A, 0x00, 0x0D, 0x0A, 0x51, 0x55, 0x49, 0x54, 0x0A,
    ]) {
        // V2 - Read header length bytes after the first 16 at [14..16]

        let mut header = [0u8; 16];
        stream.read_exact(&mut header).await?;
        let len = u16::from_be_bytes([header[14], header[15]]);

        let mut buf = vec![0u8; 16 + len as usize];
        let buf_slice = buf.as_mut_slice();
        buf_slice[..16].copy_from_slice(&header);
        stream.read_exact(&mut buf_slice[16..]).await?;
        let mut buf = buf.as_slice();

        Ok(Some(proxy_protocol::parse(&mut buf)?))
    } else {
        Ok(None)
    }
}

/// Try to parse and apply a proxy protocol header.
///
/// * If the header is not found, then no data is read from the stream and no
///     action is taken.
/// * If the header is found: `remote_addr` will be swapped with the header's
///     source addr.
/// * If the header is found, but contains invalid data: no action is taken.
/// * If the header is found, but could not be parsed: an error is returned and
///     the connection should be closed.
async fn handle_proxy_protocol(
    stream: &mut TcpStream,
    remote_addr: &mut SocketAddr,
) -> anyhow::Result<()> {
    use proxy_protocol::{
        version1::ProxyAddresses as ProxyAddrV1,
        version2::ProxyAddresses as ProxyAddrV2,
    };
    // proxy-protocol doesn't support async streams, so in order to parse the
    // proxy protocol, we need to pre-read the header before passing it in
    // without reading into tls data.
    //
    // First, check if there is a proxy protocol header with the first 12 bytes.
    // Second, read the whole proxy protocol header.
    // Third, pass the raw data into proxy-protocol to get parsed.

    match parse_proxy_protocol(stream).await? {
        Some(ProxyHeader::Version1 { addresses }) => match addresses {
            ProxyAddrV1::Ipv4 {
                source,
                destination: _,
            } => *remote_addr = SocketAddr::V4(source),
            ProxyAddrV1::Ipv6 {
                source,
                destination: _,
            } => *remote_addr = SocketAddr::V6(source),
            _ => {}
        },
        Some(ProxyHeader::Version2 {
            command: _,
            transport_protocol: _,
            addresses,
        }) => match addresses {
            ProxyAddrV2::Ipv4 {
                source,
                destination: _,
            } => *remote_addr = SocketAddr::V4(source),
            ProxyAddrV2::Ipv6 {
                source,
                destination: _,
            } => *remote_addr = SocketAddr::V6(source),
            _ => {}
        },
        _ => {}
    }
    Ok(())
}

// This is a separate function for ease of error handling
async fn start_task(
    mut stream: TcpStream,
    acceptor: TlsAcceptor,
    mut remote_addr: SocketAddr,
    server_port: u16,
) {
    use anyhow::Context;

    if CONF.proxy_protocol {
        if let Err(e) = handle_proxy_protocol(&mut stream, &mut remote_addr)
            .await
            .context("Error parsing proxy protocol")
        {
            debug!("{:#}", e);
            return;
        }
    }

    // Accept tls connection
    match acceptor
        .accept(stream)
        .await
        .context("Error accepting tls connection")
    {
        Ok(mut stream) => {
            // Handle connection
            match handle_connection(&mut stream, remote_addr, server_port).await
            {
                Ok(log_info) => log_info.log(),
                Err(ErrorLogInfo { log_info, error }) => {
                    error.log();
                    if let Some(log_info) = log_info {
                        log_info.log();
                    }
                    let mut error_output = Vec::with_capacity(1040);
                    error_output.extend_from_slice(
                        error.status().to_string().as_bytes(),
                    );
                    error_output.push(b' ');
                    error_output.extend_from_slice(error.meta().as_bytes());
                    error_output.push(b'\r');
                    error_output.push(b'\n');
                    if let Err(e) = stream
                        .write_all(&error_output)
                        .await
                        .context("Error writing response")
                    {
                        debug!("{:#}", e);
                    }
                }
            }
            if let Err(e) = stream.close().await.context("Error closing stream")
            {
                debug!("{:#}", e);
            }
        }
        Err(e) => {
            debug!("{:#}", e);
        }
    };
}

/// Parse the request, route it, call the response handler
async fn handle_connection(
    stream: &mut TlsStream<TcpStream>,
    remote_addr: SocketAddr,
    server_port: u16,
) -> std::result::Result<FullLogInfo, ErrorLogInfo> {
    use anyhow::Context;

    // Read the request
    let (input_buf, read_len) = if CONF.request_timeout == 0 {
        read_request(stream).await
    } else {
        read_request(stream)
            .or(async {
                Timer::after(Duration::from_secs(CONF.request_timeout)).await;
                Err(GemError::TimeOut)
            })
            .await
    }?;

    let (route, req) = route(
        str::from_utf8(&input_buf[..read_len])
            .context("url is not valid utf8")
            .map_err(GemError::BadReq)?,
    )?;
    let log_ip = match CONF.ip_log {
        config::IpLogAmount::Full => remote_addr.ip(),
        config::IpLogAmount::Partial => anon_ip(remote_addr.ip()),
        config::IpLogAmount::None => Ipv4Addr::new(0, 0, 0, 0).into(),
    };
    match handle_response(stream, remote_addr, server_port, route, req.clone())
        .await
    {
        Ok(log_info) => Ok(FullLogInfo {
            log_info,
            req,
            log_ip,
        }),
        Err(error) => {
            let status = error.status();
            let meta = error.meta().as_bytes().to_owned();
            Err(ErrorLogInfo {
                error,
                log_info: Some(FullLogInfo {
                    log_info: LogInfo {
                        size: 0,
                        status,
                        meta,
                    },
                    req,
                    log_ip,
                }),
            })
        }
    }
}

/// Handle responding to the request
///
/// Depending on the path requested this could mean:
/// - Return file contents with correct mime type
/// - Return a gemtext page list directory contents
/// - Run a CGI script
/// - Send a request to an SCGI script
/// - Send a redirect
/// - Return an error
async fn handle_response(
    stream: &mut TlsStream<TcpStream>,
    remote_addr: SocketAddr,
    server_port: u16,
    route: &'static Route,
    req: Request,
) -> Result<LogInfo> {
    use crate::router::RouteType;
    use anyhow::Context;

    // Check client cert if one if configured
    if let Some(config_client_cert) = &route.client_cert {
        let (_, session) = stream.get_ref();
        match session.peer_certificates().and_then(|list| list.first()) {
            Some(session_client_cert) => {
                match x509_parser::parse_x509_certificate(
                    session_client_cert.as_ref(),
                ) {
                    Ok((_, cert)) => {
                        if cert != *config_client_cert {
                            return Err(GemError::ClientCertNoAuth);
                        }
                    }
                    Err(e) => {
                        log::debug!("Bad client cert: {:#}", e);
                        return Err(GemError::BadReq(anyhow::anyhow!(
                            "Bad client cert"
                        )));
                    }
                }
            }
            None => return Err(GemError::ClientCertRequired),
        }
    }

    let log_info = match &route.route_type {
        RouteType::Cgi(cgi_route) => {
            let fut = serve_cgi(
                cgi_route,
                &route.domain,
                &req,
                remote_addr.ip(),
                server_port,
                stream,
            );
            let res = match cgi_route.timeout {
                Some(timeout) => {
                    fut.or(async {
                        Timer::after(Duration::from_secs(timeout)).await;
                        Err(GemError::CGIError(anyhow::anyhow!(
                            "CGI Process timed out"
                        )))
                    })
                    .await
                }
                None => fut.await,
            };
            // Convert ServerErrors from serve_cgi to CGIErrors
            match res {
                Ok(log_info) => Ok(log_info),
                Err(GemError::ServerError(e)) => Err(GemError::CGIError(e)),
                Err(e) => Err(e),
            }?
        }
        RouteType::Scgi(scgi_route) => {
            let fut = serve_scgi(
                scgi_route,
                &route.domain,
                &req,
                &route.path,
                remote_addr.ip(),
                server_port,
                stream,
            );
            let res = match scgi_route.timeout {
                Some(timeout) => {
                    fut.or(async {
                        Timer::after(Duration::from_secs(timeout)).await;
                        Err(GemError::CGIError(anyhow::anyhow!(
                            "SCGI Request timed out"
                        )))
                    })
                    .await
                }
                None => fut.await,
            };
            // Convert ServerErrors from serve_cgi to CGIErrors
            match res {
                Ok(log_info) => Ok(log_info),
                Err(GemError::ServerError(e)) => Err(GemError::CGIError(e)),
                Err(e) => Err(e),
            }?
        }
        RouteType::Redirect(redirect_route) => {
            let mut output = Vec::with_capacity(1024);
            if redirect_route.permanent {
                output.extend_from_slice(b"31 ");
            } else {
                output.extend_from_slice(b"30 ");
            }
            let meta = if redirect_route.redirect_rewrite {
                let mut uri = uriparse::URIReference::try_from(
                    redirect_route.url.as_str(),
                )
                .with_context(|| {
                    format!(
                        "Redirect url`{}` is not a valid uri",
                        redirect_route.url
                    )
                })
                .into_server_error()?;
                uri.set_path(req.path.as_str())
                    .with_context(|| format!("Path for redirect with `redirect-rewrite` set has an invalid path: {}", req.path))
                    .into_server_error()?;
                uri.to_string()
            } else {
                redirect_route.url.clone()
            };
            output.extend_from_slice(meta.as_bytes());
            output.extend_from_slice(b"\r\n");
            stream.write_all(&output).await.into_io_error()?;
            LogInfo {
                size: 0,
                meta: meta.into_bytes(),
                status: if redirect_route.permanent { 31 } else { 30 },
            }
        }
        RouteType::Static(static_route) => {
            get_file(
                static_route,
                &req.path,
                &route.lang,
                &route.charset,
                stream,
            )
            .await?
        }
    };

    stream
        .flush()
        .await
        .context("Error flushing socket")
        .into_io_error()?;
    Ok(log_info)
}
/// Read the request
///
/// Make sure to stop and return an error if the request is too long
pub async fn read_request(
    stream: &mut TlsStream<TcpStream>,
) -> Result<([u8; 1027], usize)> {
    use error::Context;
    // buffer size is 1024 max url len + 2 for \r\n + 1 byte to detect too long requests
    let mut input_buf = [0; 1027];
    let mut total_read = 0;
    let mut found_cr = false;

    'outer: loop {
        let read = stream
            .read(&mut input_buf[total_read..])
            .await
            .context("Error reading input from client")
            .into_io_error()?;
        if read + total_read > 1026 {
            return Err(bad_req("Request too long"));
        }
        if read == 0 {
            break;
        }
        for (i, byte) in
            input_buf[total_read..total_read + read].iter().enumerate()
        {
            if *byte == b'\r' {
                found_cr = true;
            } else if found_cr && *byte == b'\n' {
                total_read += i + 1;
                break 'outer;
            } else {
                found_cr = false;
            }
        }
        total_read += read;
    }
    Ok((input_buf, total_read.saturating_sub(2)))
}

/// Anonymize client IP address
fn anon_ip(addr: IpAddr) -> IpAddr {
    match addr {
        IpAddr::V4(addr) => {
            let [a, b, _, _] = addr.octets();
            Ipv4Addr::new(a, b, 0, 0).into()
        }
        IpAddr::V6(addr) => {
            let [a, b, c, _, _, _, _, _] = addr.segments();
            Ipv6Addr::new(a, b, c, 0, 0, 0, 0, 0).into()
        }
    }
}

pub struct LogInfo {
    size: usize,
    status: u8,
    meta: Vec<u8>,
}

struct FullLogInfo {
    log_info: LogInfo,
    log_ip: IpAddr,
    req: Request,
}

impl FullLogInfo {
    fn log(&self) {
        log::info!(
            "{}\t{}\t{}\t{}\t{}\t{}",
            self.log_ip,
            self.req.host,
            self.req.path,
            self.log_info.size,
            self.log_info.status,
            std::str::from_utf8(&self.log_info.meta)
                .unwrap_or("INVALID_UNICODE")
        );
    }
}

struct ErrorLogInfo {
    log_info: Option<FullLogInfo>,
    error: GemError,
}

impl From<GemError> for ErrorLogInfo {
    fn from(error: GemError) -> Self {
        ErrorLogInfo {
            log_info: None,
            error,
        }
    }
}

#[cfg(test)]
mod test {
    use std::time::{Duration, Instant};

    use super::*;

    use async_net::SocketAddrV4;
    use proxy_protocol::{
        version1::ProxyAddresses as ProxyAddrV1,
        version2::{
            ProxyAddresses as ProxyAddrV2, ProxyCommand, ProxyTransportProtocol,
        },
        ParseError, ProxyHeader,
    };
    use rand::{Rng, RngCore};

    fn parse_proxy(body: &[u8]) -> anyhow::Result<Option<ProxyHeader>> {
        parse_proxy_len(body).0
    }
    fn parse_proxy_len(
        body: &[u8],
    ) -> (anyhow::Result<Option<ProxyHeader>>, usize) {
        async_io::block_on(async {
            let mut stream = body;
            let output = parse_proxy_protocol(&mut stream).await;
            (output, body.len() - stream.len())
        })
    }

    fn random_u8() -> u8 {
        rand::random()
    }
    #[inline]
    fn random_bool() -> bool {
        rand::random()
    }

    fn random_addr_v4() -> SocketAddrV4 {
        let mut rng = rand::thread_rng();
        let addr: (u8, u8, u8, u8) = rng.gen();
        SocketAddrV4::new(
            Ipv4Addr::new(addr.0, addr.1, addr.2, addr.3),
            rng.gen(),
        )
    }
    fn random_addr_v6() -> SocketAddrV6 {
        let mut rng = rand::thread_rng();
        let addr: (u16, u16, u16, u16, u16, u16, u16, u16) = rng.gen();
        SocketAddrV6::new(
            Ipv6Addr::new(
                addr.0, addr.1, addr.2, addr.3, addr.4, addr.5, addr.6, addr.7,
            ),
            rng.gen(),
            0,
            0,
        )
    }

    fn random_header() -> ProxyHeader {
        if random_bool() {
            let addresses = if random_bool() {
                ProxyAddrV1::Ipv4 {
                    source: random_addr_v4(),
                    destination: random_addr_v4(),
                }
            } else {
                ProxyAddrV1::Ipv6 {
                    source: random_addr_v6(),
                    destination: random_addr_v6(),
                }
            };
            ProxyHeader::Version1 { addresses }
        } else {
            let addresses = match random_u8() & 0b11 {
                0 => ProxyAddrV2::Unspec,
                1 => ProxyAddrV2::Ipv4 {
                    source: random_addr_v4(),
                    destination: random_addr_v4(),
                },
                2 => ProxyAddrV2::Ipv6 {
                    source: random_addr_v6(),
                    destination: random_addr_v6(),
                },
                3 => {
                    let mut source = [0u8; 108];
                    let mut destination = [0u8; 108];
                    let mut rng = rand::thread_rng();
                    rng.fill_bytes(&mut source);
                    rng.fill_bytes(&mut destination);
                    ProxyAddrV2::Unix {
                        source,
                        destination,
                    }
                }
                _ => unreachable!(),
            };
            ProxyHeader::Version2 {
                command: match random_bool() {
                    true => ProxyCommand::Local,
                    false => ProxyCommand::Proxy,
                },
                transport_protocol: match random_u8() & 0b11 {
                    0 => ProxyTransportProtocol::Unspec,
                    1 => ProxyTransportProtocol::Stream,
                    2 => ProxyTransportProtocol::Datagram,
                    3 => ProxyTransportProtocol::Stream,
                    _ => unreachable!(),
                },
                addresses,
            }
        }
    }

    /// Perform a simple check that v1 headers parse properly
    #[test]
    fn test_proxy_proto_v1() {
        assert_eq!(parse_proxy(b"No header here").unwrap(), None);

        let hdr = ProxyHeader::Version1 {
            addresses: ProxyAddrV1::Ipv4 {
                source: random_addr_v4(),
                destination: random_addr_v4(),
            },
        };
        let data = proxy_protocol::encode(hdr.clone()).unwrap();
        assert_eq!(parse_proxy(&data).unwrap(), Some(hdr));

        let hdr = ProxyHeader::Version1 {
            addresses: ProxyAddrV1::Ipv6 {
                source: random_addr_v6(),
                destination: random_addr_v6(),
            },
        };
        let data = proxy_protocol::encode(hdr.clone()).unwrap();
        assert_eq!(parse_proxy(&data).unwrap(), Some(hdr));
    }

    /// Perform a simple check that v2 headers parse properly
    #[test]
    fn test_proxy_proto_v2() {
        let hdr = ProxyHeader::Version2 {
            command: ProxyCommand::Proxy,
            transport_protocol: ProxyTransportProtocol::Stream,
            addresses: ProxyAddrV2::Ipv4 {
                source: random_addr_v4(),
                destination: random_addr_v4(),
            },
        };
        let data = proxy_protocol::encode(hdr.clone()).unwrap();
        assert_eq!(parse_proxy(&data).unwrap(), Some(hdr));

        let hdr = ProxyHeader::Version2 {
            command: ProxyCommand::Proxy,
            transport_protocol: ProxyTransportProtocol::Stream,
            addresses: ProxyAddrV2::Ipv6 {
                source: random_addr_v6(),
                destination: random_addr_v6(),
            },
        };
        let data = proxy_protocol::encode(hdr.clone()).unwrap();
        assert_eq!(parse_proxy(&data).unwrap(), Some(hdr));
    }

    /// Check against as many as possible inputs to make sure that all valid headers do not error
    /// and all invalid headers match the error received by proxy_protocol
    ///
    /// This will always run for `duration` seconds
    #[test]
    fn test_proxy_proto_fuzz() {
        let duration = Duration::from_secs(15);
        let deadline = Instant::now() + duration;

        let mut count = 0;
        let mut v1_count = 0;
        let mut v2_count = 0;

        let mut rng = rand::thread_rng();

        while Instant::now() < deadline {
            // Create a random header and verify that it parses correctly
            let header = random_header();
            let mut data = proxy_protocol::encode(header.clone()).unwrap();
            let len = data.len();
            let mut extra = [0u8; 64];
            rng.fill_bytes(&mut extra);
            data.extend_from_slice(&extra);
            // We add a newline here because otherwise, parser will wait indefinitely until a
            // newline.
            data.extend_from_slice(b"\n");

            let mut data = data.to_vec();
            count += 1;

            assert_eq!(parse_proxy(&data).unwrap(), Some(header));

            let is_ascii = data.starts_with(b"PROXY ");
            if is_ascii {
                v1_count += 1 + len * 8;
            } else {
                v2_count += 1 + len * 8;
            }

            // perform N permutations, corrupting the header 1 random bit at a time where N is the
            // number of bits in the header
            for _ in 0..len * 8 {
                let i = random_u8() as usize % len;
                let o = random_u8() & 0b111;
                data[i] ^= 1 << o;

                count += 1;

                let genbuf = || {
                    if is_ascii {
                        data[..len]
                            .iter()
                            .map(|c| match *c >= 0x20 && *c <= 0x7F {
                                true => (*c as char).to_string(),
                                false => format!("\\{c:X}"),
                            })
                            .collect::<Vec<String>>()
                            .join("")
                    } else {
                        data[..len]
                            .iter()
                            .map(|c| format!("{c:02X}"))
                            .collect::<Vec<_>>()
                            .join(" ")
                    }
                };

                let (output, read) = parse_proxy_len(&data);
                if let Err(e) = &output {
                    if e.downcast_ref::<io::Error>().is_some() {
                        // This is an error with the length of the data.
                        // If we reach this point, then the input is absolutely corrupted and we
                        // cannot continue.
                        continue;
                    }
                }
                if let Ok(None) = &output {
                    if read != 0 {
                        let buf = genbuf();
                        panic!("{buf}\nThere no header, but {read} bytes were read")
                    }
                }

                let mut expected_slice = data.as_slice();
                let expected = match proxy_protocol::parse(&mut expected_slice)
                {
                    Ok(output) => Ok(Some(output)),
                    Err(ParseError::NotProxyHeader) => Ok(None),
                    Err(err) => Err(err),
                };
                let expected_read = data.len() - expected_slice.len();

                // Compare the outputs of the expected behaviour and the actual behaviour.
                // There are some edge cases because of the way we parse the data
                match (output, expected) {
                    (Ok(a), Ok(b)) => {
                        if a != b {
                            let buf = genbuf();
                            assert_eq!(a, b, "{buf}");
                        }
                        if read != expected_read {
                            let buf = genbuf();
                            assert_eq!(read, expected_read, "{buf}");
                        }
                    }
                    (Err(a), Err(b)) => {
                        // This is where results may differ. What matters however is that both
                        // produce the same type of error
                        match (a.downcast::<ParseError>().unwrap(), b) {
                            (ParseError::NotProxyHeader, _) => {
                                let buf = genbuf();
                                panic!("{buf}\nIt should not be possible for a to return a NotProxyHeader error")
                            }
                            (
                                ParseError::InvalidVersion { version: a },
                                ParseError::InvalidVersion { version: b },
                            ) => {
                                let buf = genbuf();
                                assert_eq!(
                                    ParseError::InvalidVersion { version: a }
                                        .to_string(),
                                    ParseError::InvalidVersion { version: b }
                                        .to_string(),
                                    "{buf}"
                                );
                            }
                            (
                                ParseError::Version1 { source: _ },
                                ParseError::Version1 { source: _ },
                            ) => {}
                            (
                                ParseError::Version2 { source: _ },
                                ParseError::Version2 { source: _ },
                            ) => {}
                            (a, b) => {
                                let buf = genbuf();
                                assert_eq!(
                                    a.to_string(),
                                    b.to_string(),
                                    "{buf}"
                                );
                            }
                        }
                    }
                    (Err(a), b) => match a.downcast::<ParseError>().unwrap() {
                        ParseError::Version1 { source } => {
                            match source {
                                // (V1 error - unexpected eof, an okay result)
                                // This is a rare occurrence, but it is possible that a technically
                                // valid header gets read as invalid because there is a \n hidden
                                // inside the header. This is okay so should not panic.
                                proxy_protocol::version1::ParseError::UnexpectedEof => {},
                                a => {
                                    let buf = genbuf();
                                    panic!("{buf}\n{a:?} != {b:?}")
                                }
                            }
                        }
                        a => {
                            let buf = genbuf();
                            panic!("{buf}\n{a:?} != {b:?}")
                        }
                    },
                    (a, b) => {
                        let buf = genbuf();
                        panic!("{buf}\n{a:?} != {b:?}")
                    }
                }
            }
        }
        println!(
            "Checked {count} headers ({}% v1 headers and {}% v2 headers)",
            (v1_count as f64 / count as f64 * 100.0).round(),
            (v2_count as f64 / count as f64 * 100.0).round()
        )
    }
}
